#define __NO_VERSION__

#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Include files, designed to support most kernel versions 2.0.0 and later. */
#include <linux/config.h>

#if defined(CONFIG_SMP) && ! defined(__SMP__)
#define __SMP__
#endif

#if defined(MODULE) && defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
#define MODVERSIONS
/* Older kernels do not include this automatically. */
#endif
#include <linux/version.h>

#if (LINUX_VERSION_CODE < 0x20341) &&  defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>

#include <asm/io.h>
#include "linuxfet.h"
#include "vmns_cfg.h"
#include "vmns_drv.h"

typedef 
VMNS_DRV_STATUS (*VMNS_IOCTL_FUNCTION) (struct net_device*, PVMNS_DRV_SIOC_HEADER);

#define DECLARE_FUNC(x) \
	static VMNS_DRV_STATUS x(struct net_device*,PVMNS_DRV_SIOC_HEADER)

DECLARE_FUNC(vmns_drv_identify);
DECLARE_FUNC(vmns_drv_commit);
DECLARE_FUNC(vmns_drv_disconnect);
DECLARE_FUNC(vmns_drv_exit_vmns);
DECLARE_FUNC(vmns_drv_set_callback);
DECLARE_FUNC(vmns_drv_query_status);
DECLARE_FUNC(vmns_drv_set_status);
DECLARE_FUNC(vmns_drv_set_mcam);
DECLARE_FUNC(vmns_drv_get_mcam);
DECLARE_FUNC(vmns_drv_set_vcam);
DECLARE_FUNC(vmns_drv_set_mode);
DECLARE_FUNC(vmns_drv_get_mode);
DECLARE_FUNC(vmns_drv_get_vcam);
#ifdef vmns_drv_read_reg
DECLARE_FUNC(vmns_drv_read_reg);
#endif

static
struct {
	VMNS_DRV_OPCODE			opcode;
	VMNS_IOCTL_FUNCTION 	function;	
}   func_table[]	= 	
{
	{VMNS_DRV_OPCODE_IDENTIFY,  		vmns_drv_identify},
	{VMNS_DRV_OPCODE_COMMIT,			vmns_drv_commit},
	{VMNS_DRV_OPCODE_DISCONNECT,		vmns_drv_disconnect},
	{VMNS_DRV_OPCODE_VMNS_EXIT,			vmns_drv_exit_vmns},
	{VMNS_DRV_OPCODE_SET_CALLBACK,		vmns_drv_set_callback},
	{VMNS_DRV_OPCODE_SET_MCAM,			vmns_drv_set_mcam},
	{VMNS_DRV_OPCODE_GET_MCAM,			vmns_drv_get_mcam},
	{VMNS_DRV_OPCODE_SET_VCAM,			vmns_drv_set_vcam},
	{VMNS_DRV_OPCODE_GET_VCAM,			vmns_drv_get_vcam},
	{VMNS_DRV_OPCODE_SET_MODE,			vmns_drv_set_mode},	
	{VMNS_DRV_OPCODE_GET_MODE,			vmns_drv_get_mode},		
	{VMNS_DRV_OPCODE_QUERY_STATUS,		vmns_drv_query_status},
	{VMNS_DRV_OPCODE_SET_STATUS,		vmns_drv_set_status},
	{VMNS_DRV_OPCODE_LAST,				NULL}
};

int vmns_process_ioctl(struct net_device* dev, PVMNS_DRV_SIOC_HEADER pParams) {
	int i;
	for (i=0;func_table[i].opcode!=VMNS_DRV_OPCODE_LAST;i++) 
		if (pParams->OpCode==func_table[i].opcode) {
			pParams->ret_status=func_table[i].function(dev,pParams);
			if (pParams->ret_status==VMNS_DRV_STATUS_OK) 
				return 0;
			else
				return -EAGAIN;				
		}	
	return -EOPNOTSUPP;	
}

#ifdef vmns_drv_read_reg
static VMNS_DRV_STATUS
vmns_drv_read_reg(struct net_device* dev, PVMNS_DRV_SIOC_HEADER pParams) {
	VMNS_REG_SIZE	size=((PVMNS_DRV_SIOC_GET_REG) pParams)->size;
	void*			data=(void*)&(((PVMNS_DRV_SIOC_GET_REG) pParams)->data);
	UINT32			offset=((PVMNS_DRV_SIOC_GET_REG) pParams)->offset;
	switch (size) {
	case VMNS_REG_BYTE:
		*((PBYTE) data)=inb(dev->base_addr+offset);
		break;
	case VMNS_REG_WORD:
		*((PWORD) data)=inw(dev->base_addr+offset);
		break;
	case VMNS_REG_DWORD:
		*((PDWORD) data)=inl(dev->base_addr+offset);
		break;
	case VMNS_REG_QWORD:
		*((PDWORD) data)=inl(dev->base_addr+offset);
		*((PDWORD) data+1)=inl(dev->base_addr+offset+4);				
		break;
	}
	return VMNS_DRV_STATUS_OK;	
}
#endif

static VMNS_DRV_STATUS 
vmns_drv_identify(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_PRIVATE pvmns_priv=GET_VMNS_PRIVATE(dev);
	netdev_private *np = (netdev_private *)dev->priv;
	
	pvmns_priv->flag|=VMNS_FLAG_CONNECTED;
	
	memcpy(((PVMNS_DRV_SIOC_IDENTIFY) hdr)->Signature2,
		VMNS_DRV_SIGNATURE,strlen(VMNS_DRV_SIGNATURE)+1);
		
	((PVMNS_DRV_SIOC_IDENTIFY) hdr)->version2.major=MAJOR_VERSION;
	((PVMNS_DRV_SIOC_IDENTIFY) hdr)->version2.minor=MINOR_VERSION;
	((PVMNS_DRV_SIOC_IDENTIFY) hdr)->revisionID = np->revision;
	
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS 
vmns_drv_exit_vmns(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_PRIVATE pvmns_priv=GET_VMNS_PRIVATE(dev);
	if (pvmns_priv->flag & VMNS_FLAG_OPENED) {
		if (dev->stop(dev)!=0) 
			return VMNS_DRV_STATUS_BUSY;
		pvmns_priv->flag &= (~VMNS_FLAG_OPENED);
		pvmns_priv->flag &= (~VMNS_FLAG_COMMITTED);
	}
	return VMNS_DRV_STATUS_OK;		
}	

static VMNS_DRV_STATUS 
vmns_drv_commit(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_PRIVATE	vpriv=GET_VMNS_PRIVATE(dev);
	
	if (vpriv->flag & VMNS_FLAG_COMMITTED)
		return VMNS_DRV_STATUS_BUSY;
	vpriv->flag|=VMNS_FLAG_COMMITTED;
	multicast_filter_limit = 28;
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS 
vmns_drv_disconnect(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_PRIVATE	vpriv=GET_VMNS_PRIVATE(dev);
	vpriv->flag=0;
	vpriv->notify=NULL;
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS 
vmns_drv_set_callback(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_PRIVATE		vpriv=GET_VMNS_PRIVATE(dev);
	PVMNS_DRV_SIOC_CALLBACK param=(PVMNS_DRV_SIOC_CALLBACK) hdr;
	vpriv->notify=param->func;	
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS 
vmns_drv_query_status(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_SIOC_QUERY_STATUS param=(PVMNS_DRV_SIOC_QUERY_STATUS) hdr;
	switch(param->status_id) {
	case QUERY_LINK_STATUS:
		if (readb(dev->base_addr + FET_MIISR) & 0x02)
			param->u.link_status=LINK_FAIL;
		else if (readb(dev->base_addr + FET_MIISR) & 0x04)
			param->u.link_status=LINK_FULL_DUPLEX;
		else
			param->u.link_status=LINK_HALF_DUPLEX;
		break;
	case QUERY_SPEED_STATUS:
		if (readb(dev->base_addr + FET_MIISR) & 0x01)
			param->u.speed_status=SPEED_10;
		else
			param->u.speed_status=SPEED_100;
		break;
	default:
		return VMNS_DRV_STATUS_FAIL;
	}
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS 
vmns_drv_set_status(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_SIOC_SET_STATUS param=(PVMNS_DRV_SIOC_SET_STATUS) hdr;
	switch(param->status_id) {
	case SET_LINK_STATUS:
		break;
	case SET_SPEED_STATUS:
		break;
	default:
		return VMNS_DRV_STATUS_FAIL;
	}	
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS
vmns_drv_set_mcam(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_SIOC_SET_MCAM param=(PVMNS_DRV_SIOC_SET_MCAM) hdr;
	unsigned long oldmask=CAM_mask_read(dev,CAMC_SELECT_MCAM);
	int i;
	
	if ((param->offset+param->num) >32 ) {
		return VMNS_DRV_STATUS_FAIL;		
	}
	
	for (i=param->offset;i<(param->offset+param->num);i++) {
		CAM_data_write(dev, CAMC_SELECT_MCAM, i, 
			&param->u.mcam[i-param->offset][0]);
		oldmask|=(1<<i);		
	}
	
	CAM_mask_write(dev, CAMC_SELECT_MCAM, (oldmask & param->cam_mask));	
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS
vmns_drv_get_mcam(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_SIOC_GET_MCAM param=(PVMNS_DRV_SIOC_GET_MCAM) hdr;
	unsigned long oldmask=CAM_mask_read(dev,CAMC_SELECT_MCAM);
	int i;	
	
	if ((param->offset+param->num) >32 ) {
		return VMNS_DRV_STATUS_FAIL;		
	}
	
	for (i=param->offset;i<(param->offset+param->num);i++)  {
		CAM_data_read(dev, CAMC_SELECT_MCAM, i, &param->u.mcam[i][0]);
		oldmask|=(1<<i);
	}	
	param->cam_mask=CAM_mask_read(dev, CAMC_SELECT_MCAM);
		
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS
vmns_drv_set_vcam(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_SIOC_SET_VCAM param=(PVMNS_DRV_SIOC_SET_VCAM) hdr;
	int i;
	unsigned long oldmask=CAM_mask_read(dev,CAMC_SELECT_VCAM);
	
	for (i=param->offset;i<(param->offset+param->num);i++)  {
		CAM_data_write(dev, CAMC_SELECT_VCAM, i, 
			(unsigned char*)&param->u.vcam[i-param->offset]);
		oldmask|=(1<<i);
	}
	CAM_mask_write(dev, CAMC_SELECT_VCAM, (oldmask & param->cam_mask));	
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS
vmns_drv_get_vcam(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_SIOC_GET_VCAM param=(PVMNS_DRV_SIOC_GET_VCAM) hdr;
	int i;	
	for (i=param->offset;i<param->offset+param->num;i++)  {

		CAM_data_read(dev, CAMC_SELECT_VCAM, i, 
			(unsigned char*)&param->u.vcam[i-param->offset]);
	}
	param->cam_mask=CAM_mask_read(dev, CAMC_SELECT_VCAM);		
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS
vmns_drv_set_mode(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_PRIVATE		vpriv=GET_VMNS_PRIVATE(dev);
	PVMNS_DRV_SIOC_SET_MODE	param=(PVMNS_DRV_SIOC_SET_MODE) hdr;
	vmns_set_mode(dev,vpriv,param->mode);
	return VMNS_DRV_STATUS_OK;
}

static VMNS_DRV_STATUS
vmns_drv_get_mode(struct net_device* dev, PVMNS_DRV_SIOC_HEADER hdr) {
	PVMNS_DRV_PRIVATE		vpriv=GET_VMNS_PRIVATE(dev);
	PVMNS_DRV_SIOC_GET_MODE	param=(PVMNS_DRV_SIOC_GET_MODE) hdr;
	param->mode=vpriv->mode;
	return VMNS_DRV_STATUS_OK;
}

void vmns_set_mode(struct net_device* dev, PVMNS_DRV_PRIVATE vpriv,
	unsigned long mode) {		
	unsigned char b;	
	vpriv->mode=mode;
	
	b=readb(dev->base_addr + 0x09);	
	if (mode & VMNS_DRV_MODE_DISAU)
		b |= 0x02;	//set DISAU;
	else 
		b &= ~0x02;	//disable DISAU;		
	writeb(b,dev->base_addr+0x09);
	
	b=readb(dev->base_addr+0x07);
	if (mode & VMNS_DRV_MODE_PQEN)
		b|= 1;
	else 
		b&=~1;
	if (mode & VMNS_DRV_MODE_RTGOPT)
		b|=0x10;
	else
		b&=~0x10;
	writeb(b,dev->base_addr+0x07);
		
	b=readb(dev->base_addr+0x06);
	if (mode & VMNS_DRV_MODE_DISAB)
		b&=~0x08;
	else
		b|=0x08;
	if (mode & VMNS_DRV_MODE_DISAM)
		b&=~0x04;
	else
		b|=0x04;
	writeb(b,dev->base_addr+0x06);
}
